#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include <gc/gc.h>

#if __SIZEOF_POINTER__ == 8

#define TAG_BITS 3
#define TAG_MASK 7

#elif __SIZEOF_POINTER__ == 4

#define TAG_BITS 2
#define TAG_MASK 3

#else

#error "Unsupported pointer size"

#endif

#define TAG_FIXNUM   0
#define TAG_POINTER  1
#define TAG_CHAR     2
#define TAG_OTHER    3

typedef char * Val;
typedef uintptr_t UInt;
typedef intptr_t  SInt;

typedef Val(*Fun)(Val[]); // type Fun ::= (Val[]) -> Val

#define PRI_SINTVAL PRIdPTR
#define PRI_UINTVAL PRIuPTR

#define HAS_TAG(val, tag) ((((UInt)(val)) & TAG_BITS) == (tag))
#define TAG_PTR(ptr) (((Val)(ptr)) + TAG_POINTER)
#define UNTAG_PTR(ptr, type) ((type) ((ptr) - TAG_POINTER))

#define INT_TO_FIXNUM(val) ((Val)( ( ((UInt)(val)) << TAG_BITS ) + TAG_FIXNUM ))
// NOTE: This relies on >> being an arithmetic shift for signed values.
#define FIXNUM_TO_INT(val) ( ((SInt)(val)) >> TAG_BITS )


#define VAL_OTHER(index) ((Val)(((index) << TAG_BITS) | TAG_OTHER ))

#define VAL_UNSPECIFIED  VAL_OTHER(0)
#define VAL_FALSE        VAL_OTHER(1)
#define VAL_TRUE         VAL_OTHER(2)
#define VAL_EMPTY_LIST   VAL_OTHER(3)

#define ALLOC_WORD_BLOB(n) TAG_PTR(GC_malloc(sizeof(Val) * (n)))
#define WORD_REF(blob, index) (UNTAG_PTR(blob, Val*)[index])

#define CHAR_TO_VAL(ch) (((Val)(((UInt)(ch)) << TAG_BITS)) + TAG_CHAR)

#define XTAG_CLOSURE CHAR_TO_VAL('p')
#define XTAG_STRING CHAR_TO_VAL('s')

#define VAL_IS_TRUE(val) ((val) != VAL_FALSE)
#define BOOLEANIZE(x)    ((x)? VAL_TRUE : VAL_FALSE)
