;; records.scm - An implementation of something very close to SRFI-9.
;; Unlike SRFI-9, this makes no attempt to make the produced records
;; disjunct from vectors.

(define-syntax define-record-type
  (syntax-rules ()
    ((_ name (constructor arg ...) predicate (field-name . field-accessors) ...)
     (begin

       (define (constructor arg ...)
         (vector 'name field-name ...))

       (define (predicate val)
         (and (vector? val)
              (> (vector-length val) 0)
              (eq? (vector-ref val 0) 'name)))

       (%define-record-accessors predicate 1 (field-name . field-accessors) ...)

       ))))

(define-syntax %define-record-accessors
  (syntax-rules ()
    ((_ predicate n) #t)

    ((_ predicate n (name getter) others ...)
     (begin
       (define (getter record)
         (if (predicate record)
             (vector-ref record n)
           (error "Wrong type for ~a: expecting ~a, got value ~a" 'getter 'predicate record)))
       (%define-record-accessors predicate (+ n 1) others ...)))

    ((_ predicate n (name getter setter) others ...)
     (begin
       (define (getter record)
         (if (predicate record)
             (vector-ref record n)
           (error "Wrong type for ~a: expecting ~a, got value ~a" 'getter 'predicate record)))
       (define (setter record val)
         (if (predicate record)
             (vector-set! record n val)
           (error "Wrong type for ~a: expecting ~a, got value ~a" 'setter 'predicate record)))
       (%define-record-accessors predicate (+ n 1) others ...)))

    ))
