(define-syntax when
  (syntax-rules ()
    ((when test body ...)
     (cond (test body ...)
           (else #f)))))

;; Dicts are key->value mappings. Ideally these should be implemented
;; with hash tables, but for now they are just an abstraction over
;; association lists, so we can change the implementation to a hash
;; table later.

(define-record-type dict (%make-dict alist) dict?
  (alist dict-alist set-dict-alist!))

(define (dict-ref dict key default)
  (cond ((assq key (dict-alist dict)) => cdr)
        (else default)))

(define (dict-set! dict key val)
  (cond ((assq key (dict-alist dict))
         => (lambda (binding)
              (set-cdr! binding val)))
        (else (set-dict-alist! dict
                               (cons (cons key val)
                                     (dict-alist dict))))))

(define (make-dict)
  (%make-dict '()))

(define (alist->dict alist)
  (%make-dict alist))

(define (dict-for-each proc dict)
  (for-each (lambda (binding)
              (proc (car binding) (cdr binding)))
            (dict-alist dict)))


(define (replace-string-template str values)
  (let ((len (string-length str)))
    (let loop ((beg 0) (i 0) (items '()))
      (cond ((>= i len) (reverse (if (= i beg)
                                     items
                                   (cons (substring str beg i) items))))
            ((eqv? (string-ref str i) #\@)
             (let* ((index (- (char->integer (string-ref str (+ i 1)))
                              (char->integer #\0)))
                    (item (list-ref values (- index 1))))
               (loop (+ i 2) (+ i 2)
                     (if (= beg i)
                         (cons item items)
                       (cons item (cons (substring str beg i) items))))))
            (else (loop beg (+ i 1) items))))))


;; with-output-to-string is not standard R5RS! We need a replacement...

(define (->string val)
  (cond ((string? val) val)
        ((number? val) (number->string val))
        ((symbol? val) (symbol->string val))
        (else (error '->string "Don't know how to convert to string" val))))

(define buffer-chunk-size 16)

(define-record-type <Buffer> (Buffer str pos chunks) Buffer?
  (str buffer-str set-buffer-str!)
  (pos buffer-pos set-buffer-pos!)
  (chunks buffer-chunks set-buffer-chunks!))

(define (make-buffer)
  (Buffer (make-string buffer-chunk-size) 0 '()))

(define (buffer-put-char! buf ch)
  (when (>= (buffer-pos buf) buffer-chunk-size)
    (set-buffer-chunks! buf (cons (buffer-str buf) (buffer-chunks buf)))
    (set-buffer-str! buf (make-string buffer-chunk-size))
    (set-buffer-pos! buf 0))
  (string-set! (buffer-str buf) (buffer-pos buf) ch)
  (set-buffer-pos! buf (+ 1 (buffer-pos buf))))

(define (buffer->string buf)
  (apply string-append
         (reverse
          (cons (substring (buffer-str buf) 0 (buffer-pos buf))
                (buffer-chunks buf)))))

(define (string-for-each proc str)
  (let ((len (string-length str)))
    (let loop ((i 0))
      (when (< i len)
        (proc (string-ref str i))
        (loop (+ i 1))))))

(define (buffer-put-datum! buf datum)
  (cond ((char? datum) (buffer-put-char! buf datum))
        (else (string-for-each (lambda (ch)
                                 (buffer-put-char! buf ch))
                               (->string datum)))))

(define (with-output-to-buffer proc)
  (let ((buf (make-buffer)))
    (proc (lambda (datum)
            (buffer-put-datum! buf datum)))
    (buffer->string buf)))
